@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2017  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2017-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -98,6 +98,8 @@ struct eap_aka_handle {
 	/* re-auth key */
 	uint8_t k_re[EAP_AKA_K_RE_LEN];
 
+	uint8_t rand[EAP_SIM_RAND_LEN];
+
 	char *kdf_in;
 
 	uint8_t *chal_pkt;
@@ -166,6 +168,20 @@ mk_error:
 	return false;
 }
 
+static void eap_aka_finish(struct eap_state *eap)
+{
+	struct eap_aka_handle *aka = eap_get_data(eap);
+	uint8_t session_id[1 + EAP_SIM_RAND_LEN + EAP_AKA_AUTN_LEN];
+
+	session_id[0] = EAP_TYPE_AKA;
+	memcpy(session_id + 1, aka->rand, EAP_SIM_RAND_LEN);
+	memcpy(session_id + 1 + EAP_SIM_RAND_LEN, aka->autn, EAP_AKA_AUTN_LEN);
+
+	eap_method_success(eap);
+	eap_set_key_material(eap, aka->msk, 32, aka->emsk, 32, NULL, 0,
+					session_id, sizeof(session_id));
+}
+
 static void check_milenage_cb(const uint8_t *res, const uint8_t *ck,
 		const uint8_t *ik, const uint8_t *auts, void *data)
 {
@@ -190,7 +206,7 @@ static void check_milenage_cb(const uint8_t *res, const uint8_t *ck,
 		pos += eap_sim_add_attribute(pos, EAP_SIM_AT_AUTS,
 				EAP_SIM_PAD_NONE, auts, EAP_AKA_AUTS_LEN);
 
-		eap_send_response(eap, aka->type, response, 24);
+		eap_method_respond(eap, response, 24);
 
 		return;
 	}
@@ -269,12 +285,10 @@ static void check_milenage_cb(const uint8_t *res, const uint8_t *ck,
 	l_free(aka->chal_pkt);
 	aka->chal_pkt = NULL;
 
-	eap_send_response(eap, aka->type, response, resp_len);
+	eap_method_respond(eap, response, resp_len);
 
 	if (!aka->protected) {
-		eap_method_success(eap);
-		eap_set_key_material(eap, aka->msk, 32, NULL, 0, NULL, 0,
-					NULL, 0);
+		eap_aka_finish(eap);
 
 		aka->state = EAP_AKA_STATE_SUCCESS;
 	}
@@ -434,6 +448,8 @@ static void handle_challenge(struct eap_state *eap, const uint8_t *pkt,
 
 	/* AKA' needs AUTN for prime derivation */
 	memcpy(aka->autn, autn, EAP_AKA_AUTN_LEN);
+	/* Keep RAND for session ID derivation */
+	memcpy(aka->rand, rand, EAP_SIM_RAND_LEN);
 
 	if (sim_auth_check_milenage(aka->auth, rand, autn, check_milenage_cb,
 			eap) < 0) {
@@ -502,9 +518,7 @@ static void handle_notification(struct eap_state *eap, const uint8_t *pkt,
 		/*
 		 * Server sent successful result indication
 		 */
-		eap_method_success(eap);
-		eap_set_key_material(eap, aka->msk, 32, NULL, 0, NULL, 0,
-					NULL, 0);
+		eap_aka_finish(eap);
 
 		/*
 		 * Build response packet
@@ -522,7 +536,7 @@ static void handle_notification(struct eap_state *eap, const uint8_t *pkt,
 			return;
 		}
 
-		eap_send_response(eap, aka->type, response, pos - response);
+		eap_method_respond(eap, response, pos - response);
 
 		aka->state = EAP_AKA_STATE_SUCCESS;
 
@@ -569,7 +583,7 @@ static void handle_identity(struct eap_state *eap, const uint8_t *pkt,
 			EAP_SIM_PAD_LENGTH, (uint8_t *)aka->identity,
 			strlen(aka->identity));
 
-	eap_send_response(eap, aka->type, response, pos - response);
+	eap_method_respond(eap, response, pos - response);
 }
 
 static void eap_aka_handle_request(struct eap_state *eap,
