@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2013-2015  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2013-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -41,6 +41,7 @@ const char *network_get_path(const struct network *network);
 enum security network_get_security(const struct network *network);
 const uint8_t *network_get_psk(struct network *network);
 const char *network_get_passphrase(const struct network *network);
+bool network_set_passphrase(struct network *network, const char *passphrase);
 struct l_queue *network_get_secrets(const struct network *network);
 int network_get_signal_strength(const struct network *network);
 struct l_settings *network_get_settings(const struct network *network);
@@ -52,7 +53,7 @@ const struct network_info *network_get_info(const struct network *network);
 void network_set_info(struct network *network, struct network_info *info);
 
 int network_autoconnect(struct network *network, struct scan_bss *bss);
-void network_connect_failed(struct network *network);
+void network_connect_failed(struct network *network, bool in_handshake);
 bool network_bss_add(struct network *network, struct scan_bss *bss);
 bool network_bss_list_isempty(struct network *network);
 void network_bss_list_clear(struct network *network);
@@ -69,7 +70,7 @@ int network_rank_compare(const void *a, const void *b, void *user);
 void network_rank_update(struct network *network, bool connected);
 
 void network_connect_new_hidden_network(struct network *network,
-						struct l_dbus_message *message);
+					struct l_dbus_message **message);
 
 void network_blacklist_add(struct network *network, struct scan_bss *bss);
 
