@@ -1736,17 +1736,17 @@ static void wsc_test_build_m8(const void *data)
 }
 
 static const unsigned char m8_encrypted_settings[] = {
-	0x10, 0x0e, 0x00, 0x6f, 0x10, 0x26, 0x00, 0x01, 0x00, 0x10, 0x45, 0x00,
+	0x10, 0x0e, 0x00, 0x6a, 0x10, 0x26, 0x00, 0x01, 0x01, 0x10, 0x45, 0x00,
 	0x07, 0x54, 0x65, 0x73, 0x74, 0x57, 0x50, 0x41, 0x10, 0x03, 0x00, 0x02,
-	0x00, 0x20, 0x10, 0x0f, 0x00, 0x02, 0x00, 0x08, 0x10, 0x28, 0x00, 0x01,
-	0x00, 0x10, 0x27, 0x00, 0x40, 0x34, 0x36, 0x30, 0x34, 0x44, 0x30, 0x31,
-	0x46, 0x46, 0x44, 0x42, 0x30, 0x42, 0x32, 0x39, 0x32, 0x45, 0x33, 0x37,
-	0x37, 0x33, 0x32, 0x44, 0x44, 0x34, 0x45, 0x31, 0x31, 0x43, 0x32, 0x34,
-	0x30, 0x31, 0x31, 0x35, 0x34, 0x32, 0x38, 0x39, 0x41, 0x30, 0x39, 0x41,
-	0x33, 0x33, 0x41, 0x44, 0x37, 0x30, 0x34, 0x31, 0x37, 0x37, 0x41, 0x42,
-	0x30, 0x44, 0x31, 0x42, 0x37, 0x35, 0x38, 0x44, 0x30, 0x10, 0x20, 0x00,
-	0x06, 0xa0, 0xa8, 0xcd, 0x1c, 0x7e, 0xc9, 0x10, 0x1e, 0x00, 0x08, 0xe8,
-	0x3b, 0x3b, 0xe7, 0x9e, 0x72, 0x06, 0x46,
+	0x00, 0x20, 0x10, 0x0f, 0x00, 0x02, 0x00, 0x08, 0x10, 0x27, 0x00, 0x40,
+	0x34, 0x36, 0x30, 0x34, 0x44, 0x30, 0x31, 0x46, 0x46, 0x44, 0x42, 0x30,
+	0x42, 0x32, 0x39, 0x32, 0x45, 0x33, 0x37, 0x37, 0x33, 0x32, 0x44, 0x44,
+	0x34, 0x45, 0x31, 0x31, 0x43, 0x32, 0x34, 0x30, 0x31, 0x31, 0x35, 0x34,
+	0x32, 0x38, 0x39, 0x41, 0x30, 0x39, 0x41, 0x33, 0x33, 0x41, 0x44, 0x37,
+	0x30, 0x34, 0x31, 0x37, 0x37, 0x41, 0x42, 0x30, 0x44, 0x31, 0x42, 0x37,
+	0x35, 0x38, 0x44, 0x30, 0x10, 0x20, 0x00, 0x06, 0xa0, 0xa8, 0xcd, 0x1c,
+	0x7e, 0xc9, 0x10, 0x1e, 0x00, 0x08, 0xe8, 0x3b, 0x3b, 0xe7, 0x9e, 0x72,
+	0x06, 0x46,
 };
 
 struct wsc_credential creds_1[1] = {
@@ -1831,6 +1831,22 @@ static void wsc_test_parse_m8_encrypted_settings(const void *data)
 	assert(!memcmp(expected->authenticator, m8es.authenticator, 8));
 }
 
+static void wsc_test_build_m8_encrypted_settings(const void *data)
+{
+	const struct m8_encrypted_settings_data *test = data;
+	uint8_t *out;
+	size_t out_len;
+
+	out = wsc_build_m8_encrypted_settings(&test->expected, test->creds,
+						test->n_creds, &out_len);
+	assert(out);
+
+	assert(out_len == test->len);
+	assert(!memcmp(test->pdu, out, test->len));
+
+	l_free(out);
+}
+
 static const unsigned char eap_wsc_done[] = {
 	0x01, 0x00, 0x00, 0x4a, 0x02, 0xab, 0x00, 0x4a, 0xfe, 0x00, 0x37, 0x2a,
 	0x00, 0x00, 0x00, 0x01, 0x05, 0x00, 0x10, 0x4a, 0x00, 0x01, 0x10, 0x10,
@@ -2245,6 +2261,337 @@ static void wsc_test_pin_generate(const void *data)
 	}
 }
 
+struct test_ap_sta_data {
+	struct handshake_state *ap_hs;
+	struct handshake_state *sta_hs;
+	const uint8_t ap_address[6];
+	const uint8_t sta_address[6];
+
+	uint8_t to_sta_data[1024];
+	int to_sta_data_len;
+	int to_sta_msg_cnt;
+	int to_ap_msg_cnt;
+	struct eapol_sm *ap_sm;
+	struct eapol_sm *sta_sm;
+};
+
+static int test_ap_sta_eapol_tx(uint32_t ifindex,
+					const uint8_t *dest, uint16_t proto,
+					const struct eapol_frame *ef,
+					bool noencrypt, void *user_data)
+{
+	struct test_ap_sta_data *s = user_data;
+	size_t len = sizeof(struct eapol_header) +
+		L_BE16_TO_CPU(ef->header.packet_len);
+
+	assert(ifindex == s->ap_hs->ifindex || ifindex == s->sta_hs->ifindex);
+	assert(proto == ETH_P_PAE && !noencrypt);
+
+	if (ifindex == s->ap_hs->ifindex) {	/* From AP to STA */
+		assert(!memcmp(dest, s->sta_address, 6));
+		assert(len < sizeof(s->to_sta_data));
+		memcpy(s->to_sta_data, ef, len);
+		s->to_sta_data_len = len;
+		s->to_sta_msg_cnt++;
+		return 0;
+	}
+
+	/* From STA to AP */
+	assert(!memcmp(dest, s->ap_address, 6));
+	s->to_ap_msg_cnt++;
+	__eapol_rx_packet(1, s->sta_address, proto, (const void *) ef, len,
+				noencrypt);
+	return 0;
+}
+
+static void test_ap_sta_run(struct test_ap_sta_data *s)
+{
+	eap_init();
+	eapol_init();
+	__eapol_set_tx_packet_func(test_ap_sta_eapol_tx);
+	__eapol_set_tx_user_data(s);
+
+	s->to_sta_msg_cnt = 0;
+	s->to_ap_msg_cnt = 0;
+	s->to_sta_data_len = 0;
+
+	s->ap_sm = eapol_sm_new(s->ap_hs);
+	eapol_register(s->ap_sm);
+
+	s->sta_sm = eapol_sm_new(s->sta_hs);
+	eapol_register(s->sta_sm);
+
+	eapol_start(s->sta_sm);
+	eapol_start(s->ap_sm);
+
+	while (s->to_sta_data_len) {
+		int len = s->to_sta_data_len;
+		s->to_sta_data_len = 0;
+		__eapol_rx_packet(s->sta_hs->ifindex, s->ap_address, ETH_P_PAE,
+					s->to_sta_data, len, false);
+	}
+
+	if (s->ap_sm)
+		eapol_sm_free(s->ap_sm);
+
+	if (s->sta_sm)
+		eapol_sm_free(s->sta_sm);
+
+	eapol_exit();
+	eap_exit();
+}
+
+struct test_ap_sta_hs {
+	struct handshake_state super;
+	struct test_ap_sta_data *s;
+};
+
+static struct handshake_state *test_ap_sta_hs_new(struct test_ap_sta_data *s,
+							uint32_t ifindex)
+{
+	struct test_ap_sta_hs *ths = l_new(struct test_ap_sta_hs, 1);
+
+	ths->super.ifindex = ifindex;
+	ths->super.free = (void (*)(struct handshake_state *s)) l_free;
+	ths->s = s;
+
+	return &ths->super;
+}
+
+static bool random_nonce(uint8_t nonce[])
+{
+	return l_getrandom(nonce, 32);
+}
+
+static void test_ap_sta_install_tk(struct handshake_state *hs,
+					const uint8_t *tk, uint32_t cipher)
+{
+	assert(false);
+}
+
+struct wsc_r_test_success_data {
+	bool credentials_obtained;
+	bool credentials_sent;
+	bool ap_eap_failed;
+	bool sta_eap_failed;
+	struct test_ap_sta_data *s;
+	struct wsc_credential expected_creds;
+};
+
+static void test_ap_sta_hs_event_ap(struct handshake_state *hs,
+				enum handshake_event event,
+				void *user_data, ...)
+{
+	struct wsc_r_test_success_data *data = user_data;
+	va_list args;
+
+	va_start(args, user_data);
+
+	switch (event) {
+	case HANDSHAKE_EVENT_EAP_NOTIFY:
+		assert(va_arg(args, unsigned int) ==
+			EAP_WSC_EVENT_CREDENTIAL_SENT);
+		assert(!data->credentials_sent);
+		assert(!data->ap_eap_failed);
+		data->credentials_sent = true;
+		break;
+	case HANDSHAKE_EVENT_FAILED:
+		assert(va_arg(args, int) ==
+			MMPDU_REASON_CODE_IEEE8021X_FAILED);
+		assert(!data->ap_eap_failed);
+		data->ap_eap_failed = true;
+		data->s->ap_sm = NULL;
+		break;
+	default:
+		break;
+	}
+
+	va_end(args);
+}
+
+static void test_ap_sta_hs_event_sta(struct handshake_state *hs,
+				enum handshake_event event,
+				void *user_data, ...)
+{
+	struct wsc_r_test_success_data *data = user_data;
+	va_list args;
+
+	va_start(args, user_data);
+
+	switch (event) {
+	case HANDSHAKE_EVENT_EAP_NOTIFY:
+	{
+		unsigned int eap_event = va_arg(args, unsigned int);
+		const struct wsc_credential *cred;
+
+		assert(eap_event == EAP_WSC_EVENT_CREDENTIAL_OBTAINED);
+		cred = va_arg(args, const struct wsc_credential *);
+		assert(!memcmp(cred, &data->expected_creds, sizeof(*cred)));
+		assert(!data->credentials_obtained);
+		assert(!data->sta_eap_failed);
+		data->credentials_obtained = true;
+		break;
+	}
+	case HANDSHAKE_EVENT_FAILED:
+		assert(va_arg(args, int) ==
+			MMPDU_REASON_CODE_IEEE8021X_FAILED);
+		assert(!data->sta_eap_failed);
+		data->sta_eap_failed = true;
+		data->s->sta_sm = NULL;
+		break;
+	default:
+		break;
+	}
+
+	va_end(args);
+}
+
+static void wsc_r_test_pbc_handshake(const void *data)
+{
+	static const unsigned char ap_rsne[] = {
+		0x30, 0x12, 0x01, 0x00, 0x00, 0x0f, 0xac, 0x04,
+		0x01, 0x00, 0x00, 0x0f, 0xac, 0x04, 0x01, 0x00,
+		0x00, 0x0f, 0xac, 0x02 };
+	static const char *ssid = "TestWPA2EAP";
+	static const char *ap_8021x_str = "[Security]\n"
+		"EAP-Method=WSC-R\n"
+		"[WSC]\n"
+		"UUID-R=00112233445566778899aabbccddeeff\n"
+		"RFBand=1\n"
+		"ConfigurationMethods=0x680\n";
+	static const char *sta_8021x_str = "[Security]\n"
+		"EAP-Identity=WFA-SimpleConfig-Enrollee-1-0\n"
+		"EAP-Method=WSC\n"
+		"[WSC]\n"
+		"RFBand=1\n"
+		"ConfigurationMethods=0x680\n";
+	struct l_settings *ap_8021x_settings = l_settings_new();
+	struct l_settings *sta_8021x_settings = l_settings_new();
+	struct test_ap_sta_data s = {
+		.ap_hs = test_ap_sta_hs_new(&s, 1),
+		.sta_hs = test_ap_sta_hs_new(&s, 2),
+		.ap_address = { 0x02, 0x03, 0x04, 0x05, 0x06, 0x07 },
+		.sta_address = { 0x02, 0x03, 0x04, 0x05, 0x06, 0x08 },
+	};
+	const struct wsc_credential *expected_creds = data;
+	struct wsc_r_test_success_data wsc_data = {
+		.s = &s,
+		.expected_creds = *expected_creds,
+	};
+	uint8_t uuid_e[16];
+	L_AUTO_FREE_VAR(char *, uuid_e_str) = NULL;
+	char ssid_str[33];
+
+	wsc_uuid_from_addr(s.sta_address, uuid_e);
+	uuid_e_str = l_util_hexstring(uuid_e, 16);
+
+	memcpy(wsc_data.expected_creds.addr, s.sta_address, 6);
+	memset(wsc_data.expected_creds.ssid + expected_creds->ssid_len, 0,
+		sizeof(expected_creds->ssid) - expected_creds->ssid_len);
+	memset(wsc_data.expected_creds.network_key +
+		expected_creds->network_key_len, 0,
+		sizeof(expected_creds->network_key) -
+		expected_creds->network_key_len);
+
+	memcpy(ssid_str, expected_creds->ssid, expected_creds->ssid_len);
+	ssid_str[expected_creds->ssid_len] = '\0';
+
+	l_settings_load_from_data(ap_8021x_settings, ap_8021x_str,
+					strlen(ap_8021x_str));
+	l_settings_set_string(ap_8021x_settings, "WSC", "EnrolleeMAC",
+				util_address_to_string(s.sta_address));
+	l_settings_set_string(ap_8021x_settings, "WSC", "UUID-E", uuid_e_str);
+
+	if (expected_creds->auth_type == WSC_AUTHENTICATION_TYPE_WPA2_PERSONAL) {
+		l_settings_set_string(ap_8021x_settings, "WSC",
+					"WPA2-SSID", ssid_str);
+
+		if (expected_creds->network_key_len < 64)
+			l_settings_set_string(ap_8021x_settings, "WSC",
+					"WPA2-Passphrase", (char *)
+					wsc_data.expected_creds.network_key);
+		else {
+			char psk_str[65];
+
+			/*
+			 * The PSK is 32-byte long but the setting value is
+			 * a hexstring, same as what we expect to receive in
+			 * the raw wsc_credential.network_key data.
+			 */
+			memcpy(psk_str, expected_creds->network_key, 64);
+			psk_str[64] = '\0';
+			l_settings_set_string(ap_8021x_settings, "WSC",
+						"WPA2-PSK", psk_str);
+		}
+	} else
+		l_settings_set_string(ap_8021x_settings, "WSC",
+					"Open-SSID", ssid_str);
+
+	__handshake_set_get_nonce_func(random_nonce);
+	__handshake_set_install_tk_func(test_ap_sta_install_tk);
+	__handshake_set_install_gtk_func(NULL);
+
+	handshake_state_set_authenticator(s.ap_hs, true);
+	handshake_state_set_event_func(s.ap_hs, test_ap_sta_hs_event_ap,
+					&wsc_data);
+	handshake_state_set_authenticator_address(s.ap_hs, s.ap_address);
+	handshake_state_set_supplicant_address(s.ap_hs, s.sta_address);
+	handshake_state_set_authenticator_ie(s.ap_hs, ap_rsne);
+	handshake_state_set_ssid(s.ap_hs, (void *) ssid, strlen(ssid));
+	handshake_state_set_8021x_config(s.ap_hs, ap_8021x_settings);
+
+	handshake_state_set_authenticator(s.sta_hs, false);
+	handshake_state_set_event_func(s.sta_hs, test_ap_sta_hs_event_sta,
+					&wsc_data);
+	handshake_state_set_authenticator_address(s.sta_hs, s.ap_address);
+	handshake_state_set_supplicant_address(s.sta_hs, s.sta_address);
+	handshake_state_set_authenticator_ie(s.sta_hs, ap_rsne);
+	handshake_state_set_ssid(s.sta_hs, (void *) ssid, strlen(ssid));
+	l_settings_load_from_data(sta_8021x_settings, sta_8021x_str,
+					strlen(sta_8021x_str));
+	l_settings_set_string(sta_8021x_settings, "WSC", "EnrolleeMAC",
+				util_address_to_string(s.sta_address));
+	handshake_state_set_8021x_config(s.sta_hs, sta_8021x_settings);
+
+	test_ap_sta_run(&s);
+
+	handshake_state_free(s.ap_hs);
+	handshake_state_free(s.sta_hs);
+	__handshake_set_install_tk_func(NULL);
+	l_settings_free(ap_8021x_settings);
+	l_settings_free(sta_8021x_settings);
+
+	assert(wsc_data.sta_eap_failed && wsc_data.credentials_obtained);
+	assert(wsc_data.ap_eap_failed && wsc_data.credentials_sent);
+}
+
+struct wsc_credential wsc_r_test_wpa2_cred_passphrase = {
+	.ssid_len = 7,
+	.ssid = "thessid",
+	.auth_type = WSC_AUTHENTICATION_TYPE_WPA2_PERSONAL,
+	.network_key_len = 12,
+	.network_key = "secretsecret",
+	.encryption_type = WSC_ENCRYPTION_TYPE_AES_TKIP,
+};
+
+struct wsc_credential wsc_r_test_wpa2_cred_psk = {
+	.ssid_len = 7,
+	.ssid = "thessid",
+	.auth_type = WSC_AUTHENTICATION_TYPE_WPA2_PERSONAL,
+	.network_key_len = 64,
+	.network_key = "2aacd631143bbc2d83fabf3febed3c45"
+		"6ea9f7539a1692ffb416c41cfee0c96f",
+	.encryption_type = WSC_ENCRYPTION_TYPE_AES_TKIP,
+};
+
+struct wsc_credential wsc_r_test_open_cred = {
+	.ssid_len = 7,
+	.ssid = "thessid",
+	.auth_type = WSC_AUTHENTICATION_TYPE_OPEN,
+	.network_key_len = 0,
+	.encryption_type = WSC_ENCRYPTION_TYPE_NONE,
+};
+
 int main(int argc, char *argv[])
 {
 	l_test_init(&argc, &argv);
@@ -2330,6 +2677,9 @@ int main(int argc, char *argv[])
 	l_test_add("/wsc/parse/m8 encrypted settings 1",
 			wsc_test_parse_m8_encrypted_settings,
 			&m8_encrypted_settings_data_1);
+	l_test_add("/wsc/build/m8 encrypted settings 1",
+			wsc_test_build_m8_encrypted_settings,
+			&m8_encrypted_settings_data_1);
 
 	l_test_add("/wsc/parse/wsc_done 1", wsc_test_parse_wsc_done,
 							&wsc_done_data_1);
@@ -2355,6 +2705,16 @@ int main(int argc, char *argv[])
 	l_test_add("/wsc/retransmission/no fragmentation",
 				wsc_test_retransmission_no_fragmentation, NULL);
 
+	l_test_add("/wsc-r/handshake/PBC Handshake WPA2 passphrase test",
+				wsc_r_test_pbc_handshake,
+				&wsc_r_test_wpa2_cred_passphrase);
+	l_test_add("/wsc-r/handshake/PBC Handshake WPA2 PSK test",
+				wsc_r_test_pbc_handshake,
+				&wsc_r_test_wpa2_cred_psk);
+	l_test_add("/wsc-r/handshake/PBC Handshake Open test",
+				wsc_r_test_pbc_handshake,
+				&wsc_r_test_open_cred);
+
 done:
 	return l_test_run();
 }
